// Importing Bevy ECS components and systems
use bevy::ecs::component::Component;
use bevy::ecs::system::{Commands, Query};
use bevy::ecs::entity::Entity;
use bevy::ecs::query::Without;
use bevy::ecs::system::ResMut;

// Importing Bevy assets and rendering related components
use bevy::prelude::Mesh;
use bevy::render::mesh::{Indices, PrimitiveTopology};
use bevy::asset::Assets;
use bevy::pbr::{PbrBundle, StandardMaterial};
use bevy::render::color::Color;
use bevy::transform::components::Transform;
use bevy::utils::default;

// Importing Bevy application related components
use bevy::app::{App, Plugin, Update};

// Importing Fj-core functionalities
use fj_core::algorithms::approx::Tolerance;
use fj_core::algorithms::bounding_volume::BoundingVolume;
use fj_core::algorithms::sweep::Sweep;
use fj_core::algorithms::triangulate::Triangulate;
use fj_core::objects::{Cycle, Region, Shell, Sketch, Solid};
use fj_core::operations::{BuildCycle, BuildRegion, BuildSketch, Insert, Reverse, UpdateRegion, UpdateSketch};
use fj_core::services::Services;
use fj_core::storage::Handle as FjHandle;

// Importing Fj-interop mesh
use fj_interop::mesh::Mesh as FjMesh;

// Importing Fj-math and other standard functionalities
use fj_math::{Aabb, Point, Scalar, Vector};
use std::ops::Deref;

#[derive(Component)]
struct FjSolidWrapper{
    handle: fj_core::storage::Handle<Solid>,
}

#[derive(Component, Debug)]
pub struct FjMeshWrapper {
    pub mesh: FjMesh<Point<3>>,
    pub handle: FjHandle<Solid>,
}

#[derive(Component)]
struct FjConvertedFlag;

pub struct FjRenderPlugin;

impl Plugin for FjRenderPlugin {
    fn build(&self, app: &mut App) {
        app
            .add_systems(Update, update_fj_model_system);
    }
}


fn generate_uv_mapping(fj_mesh: &FjMesh<Point<3>>) -> Vec<[f32; 2]> {
    let mut uvs = Vec::new();

    for vertex in fj_mesh.vertices() {
        let x = vertex.coords.x.into_f32();
        let y = vertex.coords.y.into_f32();
        // Here we're using x and y coordinates as u and v
        uvs.push([x, y]);
        uvs.push([x, y]);
        uvs.push([x, y]);
        uvs.push([x, y]);
        uvs.push([x, y]);
        uvs.push([x, y]);
    }

    uvs
}

fn generate_normals(fj_mesh: &FjMesh<Point<3>>) -> Vec<[f32; 3]> {
    let mut normals = Vec::new();
    for triangle in fj_mesh.triangles() {
        let normal = triangle.inner.normal();
        normals.push([normal.x.into_f32(), normal.y.into_f32(), normal.z.into_f32()]);
        normals.push([normal.x.into_f32(), normal.y.into_f32(), normal.z.into_f32()]);
        normals.push([normal.x.into_f32(), normal.y.into_f32(), normal.z.into_f32()]);
    }
    normals
}

fn generate_positions(fj_mesh: &FjMesh<Point<3>>) -> Vec<[f32; 3]> {
    let mut positions = Vec::new();

    // Iterate through each triangle
    for triangle in fj_mesh.triangles() {
        // For each vertex index in the triangle
        for vertex_index in triangle.inner.points() {
            positions.push([vertex_index.x.into_f32(), vertex_index.y.into_f32(), vertex_index.z.into_f32()]);
        }
    }

    positions
}

fn generate_indices(fj_mesh: &FjMesh<Point<3>>) -> Vec<u32> {
    let mut num_positions = 0;

    // Count the total number of positions (3 per triangle)
    for triangle in fj_mesh.triangles() {
        num_positions += 3;
    }

    // Generate the indices [0, 1, 2, ..., num_positions-1]
    (0..num_positions as u32).collect()
}

// Need to take a solid, and return all the data needed to create a Bevy Mesh from scratch
pub fn convert_mesh(fj_mesh: &FjMesh<Point<3>>) -> Mesh {
    let mut mesh = Mesh::new(PrimitiveTopology::TriangleList);
    mesh.insert_attribute(Mesh::ATTRIBUTE_POSITION, generate_positions(&fj_mesh));
    mesh.insert_attribute(Mesh::ATTRIBUTE_NORMAL, generate_normals(&fj_mesh));
    // mesh.insert_attribute(Mesh::ATTRIBUTE_UV_0, generate_uv_mapping(&fj_mesh));
    mesh.set_indices(Some(Indices::U32(generate_indices(&fj_mesh))));
    mesh
}

fn update_fj_model_system(
    mut commands: Commands,
    query: Query<(Entity, &FjMeshWrapper), Without<FjConvertedFlag>>,
    mut meshes: ResMut<Assets<Mesh>>,
    mut materials: ResMut<Assets<StandardMaterial>>,
) {
    for (entity, solid) in &query {
        let bevy_mesh = convert_mesh(&solid.mesh);
        commands.entity(entity).insert(
        (
            PbrBundle {
                mesh: meshes.add(Mesh::from(bevy_mesh.clone())),
                material: materials.add(Color::rgb(0.8, 0.7, 0.6).into()),
                transform: Transform::from_xyz(0.0, 0.0, 0.0),
                ..default()
            },
            FjConvertedFlag

        ));
    }
}




